function [Psi,J,info] = jspec(Pi)
% function [Psi,J,info] = jspec(Pi)
%
% This function constructs a J-spectral factorization for Pi. In
% other words, it factorizes Pi as:
%         Pi = Psi' J Psi
% where Psi is stable with a stable inverse and J = blkdiag(I,-I).
%
% Input
%  Pi:  Parahermitian IQC multiplier as a TF/SS object.
%
% Outputs
%  Psi: Auxiliary system as an SS object. Psi is stable and has
%          a stable inverse.
%  J: Constant matrix of the form blkdiag(I,-I)
%  info: Data structure containing fields are A, B, Q, S, R, and X.
%    In continuous time, this data yields a preliminary factorization:
%          Pi = [inv(sI-A) B; I]' [Q S; S' R] [inv(sI-A) B; I]
%    where A is Hurwitz (stable).  This is a non-square factorization.
%    The solution X of a related CARE is used to construct a J-spectral
%    factorization from this preliminary factorization.
%
%    In discrete-time, the data represents a preliminary factorization
%    of the same form as above with the Laplace variable "s" replaced by
%    "z". In this case A is a Schur (stable) matrix. X is the solution of
%    a related DARE and is used to construct the J-spectral factorization.
%
% Note: Some technical conditions on Pi are required to obtain a
% J-spectral factorization.  A sufficient condition is for the upper
% left/lower right blocks of Pi to satisfy Pi_11>0 and Pi_22<0 on the
% imaginary axis. This function assumes that the technical conditions
% required are satisfied by the multiplier Pi.

% 12/13/2013 PJS Update to IMPLICITDI for J-spectral factorizations
% 12/23/2015 PJS Update for discrete-time case

%%
% Construct minimal, state-space realization of IQC multiplier
% XXX -- "minreal" does not work on non-proper systems.
Pi = ss(Pi);
if isproper(Pi)
    Pi = minreal(Pi);
end
Ts = Pi.Ts;

if Ts==0
    %% Continuous Time Factorization
    
    %% Preliminary IQC Factorization
    % Factorize Pi in to Psi'*M*Psi where Psi is stable.
    % Use method described in Section 7.3 in "A Course in Hinf Control
    % Theory," by B. Francis.  This method is also described in the proof
    % of the first factorization theorem in the IQC paper.
    [GS,GU] = stabsep(Pi);
    nx = size(GS.A,1);
    m = size(Pi,1);
    
    Q = zeros(nx);
    R = GS.D;
    S = GS.C';
    
    %M = [Q S; S' R];
    Apsi = GS.A;
    Bpsi = GS.B;
    
    %% J-spectral Factorization
    % R = U*T*U' and T should be real, diagonal since R=R'
    % Factor R = Dpsi'*J*Dpsi where p and q are the # of positive
    % and negative evals of R, respectively.
    % XXX Check p+q = size(Pi,1)
    X = care(Apsi,Bpsi,Q,R,S);
    
    [U,T] = schur(R);
    ev = real( diag(T) );
    pidx = find( ev>0 );
    qidx = find( ev<0 );
    Dpsi = diag( sqrt([ev(pidx); -ev(qidx)]) );
    Dpsi = Dpsi*U(:,[pidx;qidx])';
        
    p = length(pidx);
    q = length(qidx);
    J = blkdiag( eye(p), -eye(q) );
    
    Cpsi = J*inv(Dpsi')*(Bpsi'*X+S');
    
    Psi = ss(Apsi,Bpsi,Cpsi,Dpsi);
else
    %% Discrete-Time Factoriztion
    
    %% Separate Stable/Proper and Unstable/non-proper parts.
    % The Matlab function "stabsep" cannot be used here since Pi
    % may be non-proper and hence in descriptor form.
    % Refs:
    %  1) Varga, SLICOTWorkingNote1998-3, p. 16
    %    (See additional refs/citations contained therein)
    %  2) Jones, Kerrigan, and Morrison, ECC, 2009.
    %  3) Kagstrom and Poromaa, Numerical Algorithms, 1996
    [Api,Bpi,Cpi,Dpi,Epi]=dssdata( Pi );
    [Api, Epi, Ptr, Qtr] = qz(Api,Epi,'real');
    [Api,Epi,Ptr,Qtr] = ordqz(Api,Epi,Ptr,Qtr,'udi');
    ee = ordeig(Api,Epi);
    idx =  abs(ee)<1;
    
    A1 = Api(idx,idx);
    A2 = Api(idx,~idx);
    A3 = Api(~idx,~idx);
    
    E1 = Epi(idx,idx);
    E2 = Epi(idx,~idx);
    E3 = Epi(~idx,~idx);
    
    % Prop 2.1 in Jones, et al
    E1i = inv(E1);
    A3i = inv(A3);
    L = dlyap(A1*E1i, E3*A3i, -(A2-A1*E1i*E2)*A3i);
    R = -E1i*(E2+L*E3);
    
    % Apply coordinate transformation
    nS = size(A1,1);
    nU = size(A3,1);
    
    BB = [eye(nS) L; zeros(nU,nS) eye(nU)]*Ptr*Bpi;
    B1 = BB(1:nS,:);
    
    CC = Cpi*Qtr*[eye(nS) R; zeros(nU,nS) eye(nU)];
    C1 = CC(:,1:nS);
    
    %% "Center" matrix of preliminary factorization
    % Note: Pi - (GS+GS') must be a constant by Liouville's theorem.
    % Thus we can compute R by evaluating this quantity at any frequency.
    % Choose a value of z0 on the unit disk to avoid poles of Pi, GS, GS'.
    Apsi = E1i*A1;
    Bpsi = E1i*B1;
    GS = ss(Apsi,Bpsi,C1,0,Ts);
    Q = zeros(nS);
    z0 = 1;
    tmp = evalfr(GS,z0);
    R = evalfr(Pi,z0) - (tmp+tmp');
    S = C1';
    %M = [Q S; S' R];
    
    %% J-spectral Factorization
    % R = U*T*U' and T should be real, diagonal since R=R'
    % Factor R = Dpsi'*J*Dpsi where p and q are the # of positive
    % and negative evals of R, respectively.
    % XXX Check p+q = size(Pi,1)
    X = dare(Apsi,Bpsi,Q,R,S);
    
    [U,T] = schur(R+Bpsi'*X*Bpsi);
    ev = real( diag(T) );
    pidx = find( ev>0 );
    qidx = find( ev<0 );
    Dpsi = diag( sqrt([ev(pidx); -ev(qidx)]) );
    Dpsi = Dpsi*U(:,[pidx;qidx])';
        
    p = length(pidx);
    q = length(qidx);
    J = blkdiag( eye(p), -eye(q) );
    
    Cpsi = J*inv(Dpsi')*(Bpsi'*X*Apsi+S');
    
    Psi = ss(Apsi,Bpsi,Cpsi,Dpsi,Ts);
end

%% Store data from preliminary factorization and ARE solution
info.A = Apsi;
info.B = Bpsi;
info.Q = Q;
info.S = S;
info.X = X;
info.R = R;